/*********************************************************************
 Author: Aliaksei Tsitovich <aliaksei.tsitovich@lu.unisi.ch>
 Roberto Bruttomesso <roberto.bruttomesso@unisi.ch>

 OpenSMT -- Copyright (C) 2009, Roberto Bruttomesso

 OpenSMT is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenSMT is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenSMT. If not, see <http://www.gnu.org/licenses/>.
 *********************************************************************/

#ifndef LRASOLVER_H
#define LRASOLVER_H

#include "TSolver.h"
#include "LAVar.h"
#include "LARow.h"
#include "LAColumn.h"
#include <glpk.h>

//
// Class to solve Linear Arithmetic theories
//
class LRASolver: public OrdinaryTSolver
{
private:

  // Structure to keep backtracking history elements
  struct LAVarHistory
  {
    Enode * e;
    LAVar * v;
    unsigned bound;
    bool bound_type;
  };

  // Possible internal states of the solver
  typedef enum
  {
    INIT, INCREMENT, SAT, UNSAT, ERROR
  } LRASolverStatus;

  typedef vector<LAVar *> VectorLAVar;

public:

  LRASolver( const int i
           , const char * n
	   , SMTConfig & c
	   , Egraph & e
	   , SStore & t
	   , vector<Enode *> & x, vector<Enode *> & d, vector<Enode *> & s) :
    OrdinaryTSolver( i, n, c, e, t, x, d, s )        // constructor
  {
    status = INIT;
    checks_history.push_back(0);
    first_update_after_backtrack = true;

    lp = glp_create_prob();
    glp_init_smcp(&glpk_parm);
    //glpk_parm.meth = GLP_DUAL;
    glpk_parm.msg_lev = GLP_MSG_OFF;
    glpk_lvars.push_back(NULL); // index 0 is not allowed by glpk
    glpk_matrix_created = false;
    glpk_bound_violated = true;
    glpk_ai.push_back(0); // index 0 is not allowed by glpk
    glpk_aj.push_back(0); // index 0 is not allowed by glpk
    glpk_cij.push_back(0); // index 0 is not allowed by glpk
    checkCount = disagreeCount = satCount = unsatCount = 0;
    variablePostFixCount = variablePostFixUnsatCount = pivotCount = 0;
    timeCheckGlpk = timeCheckOpensmt = 0;
  }
  ~LRASolver( );                                     // destructor

  lbool inform            ( Enode * );               // Inform LRA about the existence of this constraint
  bool  check             ( bool );                  // Checks the satisfiability of current constraints
  bool  assertLit         ( Enode *, bool = false ); // Push the constraint into Solver
  void  pushBacktrackPoint( );                       // Push a backtrack point
  void  popBacktrackPoint ( );                       // Backtrack to last saved point
  bool  belongsToT        ( Enode * );               // Checks if Atom belongs to this theory
  void  computeModel      ( );                       // Computes the model into enodes

protected:
  // vector in which witnesses for unsatisfiability are stored
  vector<Real> explanationCoefficients;

  VectorLAVar columns;      // Maps terms' ID to LAVar pointers
  VectorLAVar rows;         // Maps terms' ID to LAVar pointers, used to store basic columns
  VectorLAVar enode_lavar;  // Maps original constraints to solver's terms and bounds

  bool assertBoundOnColumn( LAVar * it, unsigned it_i);

  vector<unsigned> checks_history;

private:
  void doGaussianElimination( );                          // Performs Gaussian elimination of all redundant terms in the Tableau
  void update( LAVar *, const Delta & );                  // Updates the bounds after constraint pushing
  void pivotAndUpdate( LAVar *, LAVar *, const Delta &);  // Updates the tableau after constraint pushing
  void getConflictingBounds( LAVar *, vector<Enode *> & );// Returns the bounds conflicting with the actual model
  void refineBounds( );                                   // Compute the bounds for touched polynomials and deduces new bounds from it
  inline bool getStatus( );                               // Read the status of the solver in lbool
  inline bool setStatus( LRASolverStatus );               // Sets and return status of the solver
  void initSolver( );                                     // Initializes the solver
  void print( ostream & out );                            // Prints terms, current bounds and the tableau
  void addVarToRow( LAVar*, LAVar*, Real*);     //
  bool checkIntegersAndSplit();



/********** GLPK simplex stuff **********/

// Struct to order LAVar*
struct LAVar_cmp
{
  bool operator()(const LAVar *v1, const LAVar *v2) const
  {
    return v1->ID() < v2->ID();
  }
};

  static const double GLPK_EPS = 1e-6; // Experimental value of delta to handle strict inequalities
  glp_prob *lp;             // GLPK simplex structure
  glp_smcp glpk_parm;            // Parameters of the glpk simplex

  map<LAVar*, int, LAVar_cmp> glpk_var_ids; // Maps the LAVars rows to glpk row ids
  map<LAVar*, int, LAVar_cmp> glpk_row_ids; // Maps the LAVars columns to glpk columns ids
  vector<LAVar*> glpk_lvars;    // Maps the glpk ids to LAVars
  vector<int> glpk_ai, glpk_aj; // Used to create the glpk matrix: A[glpk_ai[x]][glpk_aj[x]] = glpk_cij[x]
  vector<double> glpk_cij;      // Used to create the glpk matrix: A[glpk_ai[x]][glpk_aj[x]] = glpk_cij[x]
  bool glpk_matrix_created;     // True if the simplex matrix was already created. It is done only once.
  bool glpk_bound_violated;     // True if an asserted constraint violated one of the bounds
  bool glpk_status;             // Current status of the glpk simplex

  double glpkGetDouble(const Delta &d); // Get an approximative value from the OpenSMT data structure Delta
  void glpkAddVar(LAVar *var);    // Adds a variable to the glpk simplex
  void glpkAddRow(LAVar *bvar);   // Create a empty row with bv as basic variable and add it to the glpk simplex
  void glpkAddMonomial(LAVar *bvar, LAVar *var, Real *coef); // Create a monomial "coef*var" and add it to the row where "bvar" is
  void glpkSetBounds(LAVar *var);   // Set num as the new
  int glpkCheck();             // Check the consistency of the glpk simplex

  // Debug variables
  long long checkCount;
  long long satCount;
  long long unsatCount;
  long long disagreeCount;
  long long timeCheckGlpk;
  long long timeCheckOpensmt;
  long long variablePostFixCount;
  long long variablePostFixUnsatCount;
  long long pivotCount;
  static const long long glpkTimeout = 120;  //  Timeout for the OpenSMT simplex or GLPK in seconds

  void glpkCheckTimeout();
  void glpkPrintStatistics();

/********** end of GLPK simplex stuff **********/

#ifdef PRODUCE_PROOF
  bool             fillInterpolants( );
#endif

  bool first_update_after_backtrack;

  LRASolverStatus status;                  // Internal status of the solver (different from bool)
  VectorLAVar slack_vars;              // Collect slack variables (useful for removal)
  vector<Real *> numbers_pool;             // Collect numbers (useful for removal)
  vector<LAVarHistory> pushed_constraints; // Keeps history of constraints
  set<LAVar *> touched_rows;               // Keeps the set of modified rows

  vector < LAVar * > removed_by_GaussianElimination;       // Stack of variables removed during Gaussian elimination

  // Two reloaded output operators
  inline friend ostream & operator <<( ostream & out, LRASolver & solver )
  {
    solver.print( out );
    return out;
  }

  inline friend ostream & operator <<( ostream & out, LRASolver * solver )
  {
    solver->print( out );
    return out;
  }

};

#endif
